/*
 * Code generation from interface descriptions.
 *
 * Copyright (C) 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include "client.h"
#include "common.h"
#include "config.h"
#include "declaration.h"
#include "dispatch.h"
#include "includes.h"
#include "interface.h"
#include "message.h"
#include "program.h"
#include "server.h"
#include "summary.h"
#include "templates.h"



/* Current filename details. */

char *output_basename = NULL, *output_dirname = NULL;

/* Client, server and common files. */

FILE *client_fp = NULL, *client_header_fp = NULL,
     *server_fp = NULL, *server_header_fp = NULL,
     *interface_fp = NULL;

/* Compound interface and dispatcher files. */

FILE *compound_dispatch_fp = NULL, *compound_dispatch_header_fp = NULL,
     *compound_interface_fp = NULL, *compound_interfaces_fp = NULL,
     *compound_interface_types_fp = NULL;

/* Processed interfaces. */

int processed_interfaces;

/* Maximum number of input items expected by an interface. */

int max_input_items = 0;



/* Return the first interface name as the output prefix. */

static char *get_output_prefix(struct interface *i)
{
  if (conf.output_prefix != NULL)
    return conf.output_prefix;

  if (i == NULL)
    return NULL;

  for (; i->tail != NULL; i = i->tail);
  return i->name;
}

/* Obtain the basename and directory name of generated output. */

static int set_output_location(struct interface *iface)
{
  /* Obtain a prefix for the output files. */

  conf.output_prefix = get_output_prefix(iface);

  if (conf.output_prefix == NULL)
    return 0;

  /* Obtain the basename of the output prefix. */

  output_basename = make_basename(conf.output_prefix);

  if (output_basename == NULL)
    return 0;

  /* Obtain the directory name from the output prefix. */

  if (conf.output_dir == NULL)
    output_dirname = make_dirname(conf.output_prefix);
  else
    output_dirname = conf.output_dir;

  if (output_dirname == NULL)
  {
    free(output_basename);
    return 0;
  }

  return 1;
}



/* Generate compound code output, if requested. */

int begin_compound_output(void)
{
  char *compound_dirname = conf.output_dir != NULL ? conf.output_dir : ".";
  char *s;

  if (conf.compound == NULL)
    return 1;

  processed_interfaces = 0;

  if (conf.headers)
  {
    compound_dispatch_header_fp = get_output_file(server_header_filename,
                                                  compound_dirname, conf.compound);
    if (compound_dispatch_header_fp == NULL)
      return 0;
  }

  if (conf.interfaces)
  {
    compound_interface_fp = get_output_file(interface_filename,
                                            compound_dirname, conf.compound);
    if (compound_interface_fp == NULL)
      return 0;

    compound_interfaces_fp = get_output_file(compound_interfaces_filename,
                                             compound_dirname, conf.compound);
    if (compound_interfaces_fp == NULL)
      return 0;

    compound_interface_types_fp = get_output_file(compound_interface_types_filename,
                                                  compound_dirname, conf.compound);
    if (compound_interface_types_fp == NULL)
      return 0;
  }

  if (conf.routines)
  {
    s = (conf.language == CPP_LANGUAGE) ? server_filename_cpp :
                                          server_filename_c;
    compound_dispatch_fp = get_output_file(s, compound_dirname, conf.compound);
    if (compound_dispatch_fp == NULL)
      return 0;
  }

  /* Emit prologues. */

  if (compound_dispatch_fp != NULL)
  {
    fprintf(compound_dispatch_fp, compound_dispatch_prologue, conf.compound);

    /* Write the handle function and dispatch function prologue. */

    write_handler_signature(conf.compound_name, DEFINITION_ROLE,
                            compound_dispatch_fp);
    fprintf(compound_dispatch_fp, handle_function, conf.compound_name);
    fputs(END_FUNCTION, compound_dispatch_fp);

    write_dispatcher_signature(conf.compound_name, DEFINITION_ROLE,
                               compound_dispatch_fp);
    fputs(compound_dispatch_function_prologue, compound_dispatch_fp);
  }

  if (compound_dispatch_header_fp != NULL)
  {
    fprintf(compound_dispatch_header_fp, compound_dispatch_header_prologue,
            conf.compound);

    /* Write the handler and dispatch signatures. */

    write_handler_signature(conf.compound_name, DECLARATION_ROLE,
                            compound_dispatch_header_fp);

    write_dispatcher_signature(conf.compound_name, DECLARATION_ROLE,
                               compound_dispatch_header_fp);
    fputs("\n", compound_dispatch_header_fp);
  }

  if (compound_interface_fp != NULL)
    fprintf(compound_interface_fp,
            (conf.language == CPP_LANGUAGE) ? compound_interface_prologue_cpp
                                            : compound_interface_prologue_c,
            conf.compound, conf.compound_name);

  if (compound_interfaces_fp != NULL)
    fputs(compound_interfaces_prologue, compound_interfaces_fp);

  if (compound_interface_types_fp != NULL)
  {
    fputs(compound_interface_types_prologue, compound_interface_types_fp);

    /* Include reference types. */

    fprintf(compound_interface_types_fp, ref_type_definition_prologue_c);
    fprintf(compound_interface_types_fp, ref_type_definition_epilogue_c, L4_CAP_TYPE,
            conf.compound_name);
  }

  return 1;
}

void write_compound_output(struct interface *iface)
{
  char *protocol, *ref;
  int input_items;

  if (compound_dispatch_fp != NULL)
  {
    protocol = get_protocol(iface->attributes);

    /* Populate a function dispatching to each interface-level function where
       the same protocol applies to an entire interface. */

    if (protocol != NULL)
    {
      ref = get_object_conversion(iface, 1);

      fprintf(compound_dispatch_fp, dispatch_function_interface_case,
              protocol, iface->name, ref);

      free(ref);
    }

    /* Or dispatch to each operation defined at this level. */

    else
      write_dispatcher_cases(iface->signatures, compound_dispatch_fp, iface, 1);
  }

  if (compound_dispatch_header_fp != NULL)
  {
    /* Compute the maximum number of items expected by each interface. */

    input_items = get_max_input_items(iface->signatures);

    if (input_items > max_input_items)
      max_input_items = input_items;
  }

  /* Add this interface to the compound interface. */

  if (compound_interface_fp != NULL)
    write_compound_interface(iface);

  /* Add this interface's header to the compound interface includes. */

  if (compound_interfaces_fp != NULL)
    write_include(output_basename, "_interface.h", compound_interfaces_fp);

  processed_interfaces++;
}

/* Generate includes for the server and common declarations from each input
   file. */

void write_compound_dispatch_include(void)
{
  if (compound_dispatch_header_fp != NULL)
    write_include(output_basename, "_server.h", compound_dispatch_header_fp);
}

/* Augment a compound interface class declaration. */

void write_compound_interface(struct interface *iface)
{
  char *sep;

  if (conf.language == CPP_LANGUAGE)
  {
    sep = processed_interfaces ? ", " : " ";
    fprintf(compound_interface_fp, "%spublic %s", sep, iface->name);
  }
  else
    fprintf(compound_interface_fp, "  iface_%s *to_%s;\n", iface->name, iface->name);
}

void end_compound_output(void)
{
  /* Close the files. */

  if (compound_dispatch_fp != NULL)
  {
    fputs(compound_dispatch_epilogue, compound_dispatch_fp);
    fputs(END_FUNCTION, compound_dispatch_fp);
    fclose(compound_dispatch_fp);
  }

  if (compound_dispatch_header_fp != NULL)
  {
    fprintf(compound_dispatch_header_fp, expected_items_definition,
            conf.compound_name, max_input_items);
    fclose(compound_dispatch_header_fp);
  }

  if (compound_interface_fp != NULL)
  {
    if (conf.language == CPP_LANGUAGE)
      fputs(compound_interface_epilogue_cpp, compound_interface_fp);
    else
    {
      fprintf(compound_interface_fp, compound_interface_epilogue_c,
              conf.compound_name);

      /* Include object types. */

      fprintf(compound_interface_fp, object_type_definition_c,
              conf.compound_name, conf.compound_name, conf.compound_name);
    }

    fclose(compound_interface_fp);
  }

  if (compound_interfaces_fp != NULL)
    fclose(compound_interfaces_fp);

  if (compound_interface_types_fp != NULL)
    fclose(compound_interface_types_fp);
}



/* Generate files containing interface details. */

void write_files(struct interface *iface)
{
  char *filename;

  /* NOTE: Should exit with an error. */

  if (!set_output_location(iface))
    return;

  /* Include server details for a compound interface dispatch function. */

  write_compound_dispatch_include();

  /* Open the separate files and write the details of each interface. */

  if (conf.headers)
  {
    if (conf.client)
    {
      client_header_fp = get_output_file(client_header_filename, output_dirname, output_basename);
      if (client_header_fp == NULL)
        goto finalisation;
    }

    if (conf.server)
    {
      server_header_fp = get_output_file(server_header_filename, output_dirname, output_basename);
      if (server_header_fp == NULL)
        goto finalisation;
    }
  }

  if (conf.interfaces)
  {
    interface_fp = get_output_file(interface_filename, output_dirname, output_basename);
    if (interface_fp == NULL)
      goto finalisation;
  }

  if (conf.routines)
  {
    if (conf.client)
    {
      filename = (conf.language == CPP_LANGUAGE) ? client_filename_cpp : client_filename_c;
      client_fp = get_output_file(filename, output_dirname, output_basename);
      if (client_fp == NULL)
        goto finalisation;
    }

    if (conf.server)
    {
      filename = (conf.language == CPP_LANGUAGE) ? server_filename_cpp : server_filename_c;
      server_fp = get_output_file(filename, output_dirname, output_basename);
      if (server_fp == NULL)
        goto finalisation;
    }
  }

  /* Emit prologues. */

  if (client_fp != NULL)
    fprintf(client_fp, client_prologue, output_basename, output_basename);

  if (client_header_fp != NULL)
    fputs(header_prologue, client_header_fp);

  if (server_fp != NULL)
    fprintf(server_fp, server_prologue, output_basename, output_basename);

  if (server_header_fp != NULL)
    fputs(server_header_prologue, server_header_fp);

  if (interface_fp != NULL)
    fputs(header_prologue, interface_fp);

  /* Write the interface details. */

  write_interfaces(iface);

finalisation:

  /* Close the files. */

  if (client_fp != NULL)
    fclose(client_fp);

  if (client_header_fp != NULL)
    fclose(client_header_fp);

  if (server_fp != NULL)
    fclose(server_fp);

  if (server_header_fp != NULL)
    fclose(server_header_fp);

  if (interface_fp != NULL)
    fclose(interface_fp);

  /* Free the names. */

  if (conf.output_dir == NULL)
    free(output_dirname);

  free(output_basename);
}

/* Generate the source file corresponding to the interface. */

void write_interfaces(struct interface *iface)
{
  int input_items;

  if (iface == NULL)
    return;

  /* The list is reversed. */

  write_interfaces(iface->tail);

  /* Write this particular interface after all referenced ones. */

  if (conf.verbose)
    show_interface(iface);

  /* Record details of this interface in the compound output. */

  write_compound_output(iface);

  /* Emit function definitions. */

  if (client_fp != NULL)
  {
    write_functions(iface->signatures, client_fp, iface,
                    write_client_function);

    if (conf.language == C_LANGUAGE)
      write_client_interface(iface, client_fp);
  }

  if (server_fp != NULL)
  {
    write_functions(iface->signatures, server_fp, iface,
                    write_server_function);

    /* Emit handle and dispatch functions. */

    write_handler_signature(iface->name, DEFINITION_ROLE, server_fp);
    fprintf(server_fp, handle_function, iface->name);
    fputs(END_FUNCTION, server_fp);

    write_dispatcher(iface->signatures, server_fp, iface);
  }

  /* Emit the headers. */

  if (client_header_fp != NULL)
  {
    /* Emit interface include to obtain the type employed by the client
       interface. */

    write_include(output_basename, "_interface.h", client_header_fp);

    write_interface_definition(iface, CLIENT_ROLE, client_header_fp);
  }

  if (server_header_fp != NULL)
  {
    /* Emit interface include to obtain the interface type. */

    write_include(output_basename, "_interface.h", server_header_fp);

    /* Emit signatures. */

    write_signatures(iface->signatures, DECLARATION_ROLE, server_header_fp,
                     iface, write_server_signature);

    /* Emit dispatch and handle function signatures. */

    write_dispatcher_signature(iface->name, DECLARATION_ROLE, server_header_fp);
    write_handler_signature(iface->name, DECLARATION_ROLE, server_header_fp);
  }

  /* The server interface is used by client and server code. */

  if (interface_fp != NULL)
    write_interface_definition(iface, SERVER_ROLE, interface_fp);
}

/* Generate functions corresponding to the signatures. */

void write_functions(struct signature *sig, FILE *fp, struct interface *iface,
                     void (*write_function)(struct signature *, FILE *, struct interface *))
{
  if (sig == NULL)
    return;

  write_function(sig, fp, iface);
  write_functions(sig->tail, fp, iface, write_function);
}
